/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "datebook.h"
#include "datebookday.h"
#include <datebookmonth.h>
//#include "datebookweek.h"

#include <pixmaploader.h>
#include "dateentryimpl.h"
#include <organizer.h>

#include <qtoolbar.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qaction.h>
#include <qapplication.h>
#include <qwidgetstack.h>

#include <stdlib.h>

#define DAY 1
#define WEEK 2
#define MONTH 3

static QString dateBookFilename()
{
    QString filename=getenv("HOME");
    filename += "/Documents/datebook.xml";
    return filename;
}

DateBook::DateBook( QWidget *parent, const char *name, WFlags f )
    : QMainWindow( parent, name, f )
{
    setCaption( tr("Date Book") );
    setToolBarsMovable( FALSE );

    QToolBar *bar = new QToolBar( this );
    bar->setHorizontalStretchable( TRUE );

    QMenuBar *mb = new QMenuBar( bar );

    QPopupMenu *file = new QPopupMenu( this );
    QPopupMenu *view = new QPopupMenu( this );

    bar = new QToolBar( this );

    QAction *a = new QAction( tr( "New" ), PixmapLoader::loadPixmap( "new" ), QString::null, CTRL + Key_N, this, 0 );
    connect( a, SIGNAL( activated() ), this, SLOT( fileNew() ) );
    a->addTo( file );
    a->addTo( bar );

    a = new QAction( tr( "Close" ), PixmapLoader::loadPixmap( "close" ), QString::null, CTRL + Key_W, this, 0 );
    connect( a, SIGNAL( activated() ), this, SLOT( close() ) );
    a->addTo( file );

    bar = new QToolBar( this );

    QActionGroup *g = new QActionGroup( this );
    g->setExclusive( TRUE );

    a = new QAction( tr( "Day" ), PixmapLoader::loadPixmap( "day" ), QString::null, 0, g, 0 );
    connect( a, SIGNAL( activated() ), this, SLOT( viewDay() ) );
    a->addTo( bar );
    a->addTo( view );
    a->setToggleAction( TRUE );
    a->setOn( TRUE );
    dayAction = a;
#if 0
    // datebookweek has a bug in scrolling -- disable it for the release
    a = new QAction( tr( "Week" ), PixmapLoader::loadPixmap( "week" ), QString::null, 0, g, 0 );
    connect( a, SIGNAL( activated() ), this, SLOT( viewWeek() ) );
    a->addTo( bar );
    a->addTo( view );
    a->setToggleAction( TRUE );
    weekAction = a;
#endif
    a = new QAction( tr( "Month" ), PixmapLoader::loadPixmap( "month" ), QString::null, 0, g, 0 );
    connect( a, SIGNAL( activated() ), this, SLOT( viewMonth() ) );
    a->addTo( bar );
    a->addTo( view );
    a->setToggleAction( TRUE );
    monthAction = a;

    mb->insertItem( tr( "File" ), file );
    mb->insertItem( tr( "View" ), view );

    views = new QWidgetStack( this );
    setCentralWidget( views );

    dayView = new DateBookDay( this, "day view" );
    views->addWidget( dayView, DAY );
#if 0
    weekView = new DateBookWeek( this, "week view" );
    views->addWidget( weekView, WEEK );
    connect( weekView, SIGNAL( showDate( const QDate & ) ),
	     this, SLOT( showDay( const QDate & ) ) );
    connect( dayView, SIGNAL( newEvent( const Event & ) ),
	     weekView, SLOT( addEvent( const Event & ) ) );
#endif
    monthView = new DateBookMonth( this, "month view" );
    views->addWidget( monthView, MONTH );
    viewDay();

    connect( monthView, SIGNAL( dateClicked( int, int, int ) ),
	     dayView, SLOT( setDate( int, int, int ) ) );

    load();

    connect( dayView, SIGNAL( newEvent( const Event & ) ),
	     this, SLOT( saveRequired() ) );
    connect( dayView, SIGNAL( removeEvent( const Event & ) ),
	     this, SLOT( saveRequired() ) );
    connect( dayView, SIGNAL( editEvent( const Event & ) ),
	     this, SLOT( editEvent( const Event & ) ) );

    resize( 230, 300 );
}

void DateBook::load()
{
    QFile f( dateBookFilename() );
    if ( !f.open( IO_ReadOnly ) )
	return;

    QDate hold = dayView->date();

    QTextStream ts( &f );
    QXmlInputSource inputsource( ts );
    QXmlSimpleReader reader;
    XmlHandler handler;

    reader.setFeature( "http://trolltech.com/xml/features/report-whitespace-only-CharData", FALSE );
    reader.setContentHandler( &handler );
    reader.parse( inputsource );

    Node *n = handler.firstNode();

    while ( n ) {
	if ( n->tagName() == "event" ) {
	    Event e;
	    e.load( n->firstChild() );
	    addEvent( e );
	    n = n->nextNode();
	    continue;
	}
	if ( n->lastChild() ) {
	    n = n->firstChild();
	} else {
	    if ( !n->nextNode() && n->parentNode() )
		n = n->parentNode();
	    n = n->nextNode();
	}
    }

    dayView->setDate( hold.year(), hold.month(), hold.day() );
}

void DateBook::saveRequired()
{
    // #### Could start a timer, etc.
    dayView->saveAllEvents(dateBookFilename());
}

void DateBook::fileNew()
{
    QDateTime start, end;
    if ( views->visibleWidget() == dayView ) {
	dayView->selectedDates( start, end );
    } else if ( views->visibleWidget() == monthView ) {
	QDate d = monthView->selectedDate();
	start = end = d;
	start.setTime( QTime( 10, 0 ) );
	end.setTime( QTime( 12, 0 ) );
    }
    
    DateEntry e( start, end, this );
#ifdef _WS_QWS_
    e.showMaximized();
#endif
    if ( e.exec() )
	addEvent( e.event() );
}

void DateBook::viewDay()
{
    initDay();
    views->raiseWidget( dayView );
}

void DateBook::viewWeek()
{
#if 0    
    initWeek();
    views->raiseWidget( weekView );
#endif
}

void DateBook::viewMonth()
{
    initMonth();
    views->raiseWidget( monthView );
}

void DateBook::editEvent( const Event &e )
{
    DateEntry entry( e, this );
#ifdef _WS_QWS_
    entry.showMaximized();
#endif
    if ( entry.exec() )
	addEvent(entry.event());
    else
	addEvent(e);
}

void DateBook::addEvent( const Event &e )
{
    QDate d = e.start().date();
    dayView->setDate( d.year(), d.month(), d.day() );
    dayView->addEvent( e );
}

void DateBook::showDay( const QDate &d )
{
    dayView->setDate( d.year(), d.month(), d.day() );
    views->raiseWidget( dayView );
    dayAction->setOn( TRUE );
}

void DateBook::initDay()
{
    if ( !views->visibleWidget() )
	return;
#if 0
    if ( views->visibleWidget()->inherits( "DateBookWeek" ) ) {
	QDate d = ( (DateBookWeek*)views->visibleWidget() )->date();
	dayView->setDate( d.year(), d.month(), d.day() );
    }
#endif    
}

void DateBook::initWeek()
{
#if 0    
    if ( !views->visibleWidget() )
	return;
    if ( views->visibleWidget()->inherits( "DateBookDay" ) ) {
	QDate d = ( (DateBookDay*)views->visibleWidget() )->date();
	weekView->setDate( d.year(), ( d.dayOfYear() + QDate( d.year(), 1, 1 ).dayOfWeek() - 1 ) / 7 );
    }
#endif    
}

void DateBook::initMonth()
{
}
