/**********************************************************************
** Copyright (C) 2000 Trolltech AS.  All rights reserved.
**
** This file is part of Qt Palmtop Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
#include "mpegview.h"
#include "libmpeg3.h"
#include <qfile.h>
#include <qimage.h>
#include <qpainter.h>
#include "pixmaploader.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>


static int dsp_fd;
static const int sound_fragment_size = 12;
static const int sound_fragment_bytes = 4096;
static const int sound_fragment_number = 8;
static const int bytes_per_sample = 2; //16 bit
//static const int sound_fragment_size = 14;
//static const int sound_fragment_bytes = 16384;

static int pcmVolume = -1;

#ifdef QPIM_STANDALONE
# define TIME_DEBUG
#endif


#ifdef TIME_DEBUG
static double audio_timestamp;
static double video_timestamp;
#endif

void MpegView::openAudio()
{
    
    audioSampleCounter = 0;
    audioPauseTime = 0;
#ifdef TIME_DEBUG
    qDebug( "freq = %d\n", freq );
#endif
    int fd = dsp_fd = ::open("/dev/dsp",O_WRONLY|O_NONBLOCK);
    if ( fd < 0 ) {
	// For debugging purposes - defined QT_NO_SOUND if you
	// don't have sound hardware!
	qWarning("Sending sound data to /tmp/dsp");
	fd = ::open("/tmp/dsp",O_WRONLY);
    }

    int v;
#ifdef TIME_DEBUG
    ioctl( dsp_fd, SNDCTL_DSP_GETCAPS, &v );
    if ( v & DSP_CAP_BATCH )
	qDebug( "Sound card has local buffer" );
	
    if ( v & DSP_CAP_REALTIME )
	qDebug( "Sound card has realtime sync" );

    if ( v & DSP_CAP_TRIGGER )
	qDebug( "Sound card has precise trigger" );

    if ( v & DSP_CAP_MMAP )
	qDebug( "Sound card can mmap" );
#endif
    
    v= 0x10000*sound_fragment_number+sound_fragment_size; 
    ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &v);

    //### Should use SNDCTL_DSP_GETBLKSIZE to check that it worked!
    
    v=AFMT_S16_LE; ioctl(fd, SNDCTL_DSP_SETFMT, &v);
    if ( channels == 2 ) {
	v=channels; ioctl(fd, SNDCTL_DSP_CHANNELS, &v);
	if ( v != channels )
	    qWarning( "Cannot set stereo" );
	channels = v;
    } else {
	channels = 1;
    }
    
    v = freq; ioctl(fd, SNDCTL_DSP_SPEED, &v );

    //### must check which frequency is actually used.

    
    if ( ioctl( dsp_fd, SOUND_MIXER_READ_PCM, &v ) != -1 )
	pcmVolume = v;

    
    static const int size=1;
    short int buf[size];
    /*int err = */mpeg3_read_audio( mpeg, 0, buf, 0, 
				size, stream );

    double first_timestamp = mpeg3_get_time(mpeg);
    if ( first_timestamp > 0.5 )
	audioPauseTime = int( first_timestamp*1000 );

#ifdef TIME_DEBUG
    qDebug( "Initial audio timestamp %g", first_timestamp );
    audio_timestamp = first_timestamp;
#endif    
}



/*
  Returns FALSE when the last piece of audio data has been fed to the
  device.
 */
bool MpegView::playAudio()
{
    audio_buf_info info;
    const int fudge = 40; //It takes time to start playing sound, and
    // there is time between each timer event. 
    if ( playtime.elapsed() + fudge < audioPauseTime )
	return TRUE;
    
    /* Ask OSS if there is any free space in the buffer. */
    if (ioctl(dsp_fd,SNDCTL_DSP_GETOSPACE,&info) != 0) {
	perror("Unable to query buffer space");
	::close(dsp_fd);
	exit(1);
    };

    /* Any empty fragments? */
    if (info.fragments > 0) {

	int //frags = 1;
//if ( info.fragments >= sound_fragment_number - 1 )
	frags = info.fragments;
    
	for ( int i = 0; i < frags; i++ ) {
	    int samples = sound_fragment_bytes/bytes_per_sample;
	    short int buffer[samples];
	    const int stream = 0;
	    int err;
	    if ( channels == 2 ) {
		short int left[samples/channels];
		short int right[samples/channels];
		err = mpeg3_read_audio( mpeg, 0, left, 0, 
					samples/channels, stream );
		err = mpeg3_reread_audio( mpeg, 0, right, 1, 
					  samples/channels, stream );
		for ( int i = 0; i < samples/channels; i++ ) {
		    buffer[i*channels]=left[i];
		    buffer[i*channels+1]=right[i];
		}
	    } else {
		const int channel = 0; 
		err = mpeg3_read_audio( mpeg, 0, buffer, channel, 
					samples, stream );
	    }
	    if ( err ) {
		qWarning( "Error reading audio data." );
		return FALSE;
	    }
	    int blocksize = sound_fragment_bytes; 
	    int written = write(dsp_fd,&buffer,blocksize);
	    if (written == -1) {
		perror("playAudio: error writing to sound device");
		::close(dsp_fd);
		exit(1);
	    };  
	    audioSampleCounter++;
	}
    }
    // Timestamp support in libmpeg3 is not good enough for synchronization
    int sampletime = int(double(audioSampleCounter)*(sound_fragment_bytes/bytes_per_sample/channels)/freq*1000) + audioPauseTime;
    
    static bool outputDelay = TRUE;
    
    if ( outputDelay ) {
	int buffered;
	if (ioctl(dsp_fd,SNDCTL_DSP_GETODELAY,&buffered) != 0) {
	    outputDelay = FALSE;
	    qDebug( "SNDCTL_DSP_GETODELAY not supported\n" );
	    return TRUE;
	}
	int realaudiotime = sampletime - 
			    buffered*1000/bytes_per_sample/channels/freq;
	int skew = playtime.elapsed() - realaudiotime;
    
	int corr = 0;
	if ( skew < -1000 || skew > 1000 )
	    corr = skew; 
	else 
	    corr = skew/5;

	if ( corr ) {
	    playtime = playtime.addMSecs( corr );
	}
#ifdef TIME_DEBUG
	audio_timestamp = mpeg3_get_time(mpeg);

	qDebug( "audio_timestamp %g, audiotime %d elapsed %d, skew %d, corr %d", 
		audio_timestamp, realaudiotime, playtime.elapsed(), skew, corr );
#endif
    }
    return audioSampleCounter <= total_audio_samples/sound_fragment_bytes *bytes_per_sample*channels;
}

/*
  Disgusting back-door hack: if vol < 0,  just do mute/unmute.
*/
void MpegView::setVolume( int vol )
{
    if ( dsp_fd <= 0 )
	return;
    
    int v = QMIN( QMAX(vol,0), 100);
    if ( vol >= 0 ) {
	if ( pcmVolume == v  )
	    return;
	pcmVolume = v;
    }
    v = isMuted ? 0 : pcmVolume;
    
    v = v | v << 8;
    ioctl(dsp_fd,SOUND_MIXER_WRITE_PCM,&v);
}

int MpegView::volume() const
{ 
    return pcmVolume; 
}

#define MPEG3_SEQUENCE_START_CODE 0x000001b3
#define BUFFER_SIZE 1000000

MpegView::MpegView( QWidget* parent, const char* name, WFlags fl ) :
    QWidget(parent,name,fl)
{
    isLooping = FALSE;
    isMuted = FALSE;
    mpeg = 0;
    setSizePolicy(
	QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding )
    );
}

MpegView::~MpegView()
{
    stop();
}

void MpegView::stop()
{
    if ( mpeg ) {
	if ( timerid )
	    killTimer(timerid);
	mpeg3_close( mpeg );
	if ( dsp_fd >= 0 ) {
	    ::close( dsp_fd );
	    dsp_fd = 0;
	}
	mpeg = 0;
    }
}

bool MpegView::play(const QString& filename)
{
    stop();
    dontSkip = isMuted;
    fileName = filename;
    stream = 0; // only play stream 0 for now
    current_frame = 0;
    frameskip = 0;
    total_video_frames = total_audio_samples = 0;
    
    QCString fn = QFile::encodeName(filename);

    if ( !mpeg3_check_sig(fn.data()) ) {
	mpeg = 0;
	return FALSE;
    }

    if ( !(mpeg = mpeg3_open(fn.data())) )
	return FALSE;

    
    if ( mpeg3_has_audio( mpeg ) )
       astreams = mpeg3_total_astreams( mpeg );
    else
        astreams = 0;

    if ( astreams ) {
       int astream = 0;
       channels = mpeg3_audio_channels( mpeg, astream );
       total_audio_samples = mpeg3_audio_samples( mpeg, astream );
#ifdef TIME_DEBUG
       qDebug( "%d-channel", channels );
#endif
       freq = mpeg3_sample_rate( mpeg, astream);
       openAudio();
    }

    vstreams = mpeg3_total_vstreams( mpeg );
    
    if ( vstreams ) {
	total_video_frames = totalFrames();
	framerate = mpeg3_frame_rate( mpeg, stream );
#ifdef TIME_DEBUG	
	qDebug( "Frame rate %g total %ld", framerate, total_video_frames );
#endif
	if ( framerate <= 1.0 ) {
#ifdef TIME_DEBUG	
	    qDebug( "Crazy frame rate, resetting to sensible" );
#endif
	    framerate = 25;
	}
	if ( total_video_frames == 1 ) {
#ifdef TIME_DEBUG	
	    qDebug( "Cannot seek to frame" );
#endif
	    dontSkip = TRUE;
	}

	
	setBackgroundColor( black );
    } else {
	setBackgroundColor( green.dark() );
    }
    timerid = startTimer(0);
    playtime.start();
    updateGeometry();

    return TRUE;
}

void MpegView::setCurrentFrame(long f)
{
    current_frame = f;
    frameskip = 0;
    if ( vstreams ) {
	if ( mpeg )
	    mpeg3_set_frame( mpeg, f, stream );
	playtime.start();
	playtime.addMSecs( -int(float(f)/framerate*1000) );
    }
}

void MpegView::setCurrentFrame(int f)
{
    setCurrentFrame((long)f);
}


long MpegView::totalFrames() const
{
    return mpeg ? mpeg3_video_frames( mpeg,stream) : 0;
}

void MpegView::mouseReleaseEvent(QMouseEvent* event)
{
    if ( rect().contains(event->pos()) )
	emit clicked();
}

void MpegView::timerEvent(QTimerEvent* event)
{
    if ( event->timerId() != timerid )
	return;
    killTimer(timerid);
    timerid = 0;
    bool more = FALSE;
    int delay = 0;

    if ( astreams && !isMuted ) {
	// turning off mute is not likely to work too well...
	more = playAudio( );
	delay = (sound_fragment_bytes*1000/bytes_per_sample)/freq/channels;
    } 
   
    if ( vstreams ) {
	int sw = mpeg3_video_width(mpeg,stream);
	int sh = mpeg3_video_height(mpeg,stream);

	int w = width();
	int h = height();

	if ( w*sh > sw*h ) {
	    w = sw*h/sh;
	} else {
	    h = sh*w/sw;
	}

	int dd = QPixmap::defaultDepth();

	QImage frame(w+2,h,dd==16?16:32);

	int format = dd==16 ? MPEG3_RGB565 : MPEG3_BGRA8888;

#ifdef TIME_DEBUG
	long set_frame = frameskip ? current_frame+frameskip : 0;
#endif

	
	if ( frameskip > 0 ) 
	    mpeg3_set_frame( mpeg, current_frame+frameskip, stream );

	
	if ( mpeg3_read_frame(mpeg, frame.jumpTable(),
			      0, 0, sw, sh, w, h, format, 0) == 0 )
	{
#ifdef TIME_DEBUG
	    video_timestamp = mpeg3_get_time(mpeg);
#endif
	    QPainter p(this);
	    if ( width()*sh > sw*height() ) {
		int ox = (width()-w)/2;
		p.drawImage(ox,0,frame,0,0,w,h);
	    } else {
		int oy = (height()-h)/2;
		p.drawImage(0,oy,frame,0,0,w,h);
	    }

	    current_frame = mpeg3_get_frame(mpeg,stream);
	    emit frameDone( current_frame );
	    more = TRUE;
#ifdef TIME_DEBUG
	    if ( set_frame )
		qDebug( "Set to %ld, got %ld ", set_frame, current_frame );
#endif

	    
	    
	    int frametime = int(float(current_frame+frameskip)/framerate*1000);
	    delay = frametime - playtime.elapsed();
	    
	    int offset = int(delay*framerate);

	    if ( !dontSkip &&
		 ((frameskip && offset > 1000) || offset < -1000 )) {
		if ( offset < 0 ) 
		    frameskip++;
		else
		    frameskip--;
#ifdef TIME_DEBUG
		qDebug( "new frameskip %d (offset %d)", frameskip, offset );
#endif
		//recalc delay using same code as above (###)
		frametime = int(float(current_frame+frameskip)/framerate*1000);
		delay = frametime - playtime.elapsed();
	    }
#ifdef TIME_DEBUG	    
	    qDebug ( "frametime %d, delay %d", frametime, delay );
#endif
	    if ( delay < 0 ) delay = 0;
	}
#ifdef TIME_DEBUG
	qDebug( " next video_timestamp %g   realtime %g", video_timestamp, 
		 float(playtime.elapsed())/1000);
#endif	
    } 
    if ( more ) {
	timerid = startTimer( delay );
    } else {
	if ( isLooping )
	    play( fileName );
    }
}

QSize MpegView::sizeHint() const
{
    if ( mpeg )
	return QSize(
	    mpeg3_video_width(mpeg,stream),
	    mpeg3_video_height(mpeg,stream)
	);
    else
	return QSize(40,30);
}


void MpegView::setLooping( bool on )
{
    isLooping = on;
}


void MpegView::playAgain()
{
    if ( !fileName.isEmpty() )
	play( fileName );
}



void MpegView::setMute( bool on )
{
    isMuted = on;
    setVolume( -1 ); //back door to mute/restore
    if ( isMuted ) {
	frameskip = 0;
	dontSkip = TRUE;
    }
}
